﻿Imports Microsoft.VisualBasic
Imports System
' demo L1 and L2 regularization on a logistic regression problem
Namespace Regularization
    Friend Class RegularizationProgram
        Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin L1 and L2 Regularization demo")

            Dim numFeatures As Integer = 12
            Dim numRows As Integer = 1000
            Dim seed As Integer = 42 ' interesting seeds: 28, 32, (42), 56, 58, 63, 91

            Console.WriteLine(vbCrLf + "Generating " & numRows & " artificial data items with " & numFeatures & " features")
            Dim allData()() As Double = MakeAllData(numFeatures, numRows, seed)

            Console.WriteLine("Creating train (80%) and test (20%) matrices")
            Dim trainData()() As Double = Nothing
            Dim testData()() As Double = Nothing
            MakeTrainTest(allData, 0, trainData, testData)
            Console.WriteLine("Done")

            Console.WriteLine(vbCrLf & "Training data: " & vbCrLf)
            ShowData(trainData, 4, 2, True)

            Console.WriteLine(vbCrLf & "Test data: " & vbCrLf)
            ShowData(testData, 3, 2, True)

            Console.WriteLine("Creating LR binary classifier")
            Dim lc As New LogisticClassifier(numFeatures)

            Dim maxEpochs As Integer = 1000
            Console.WriteLine(vbCrLf & "Starting training using no regularization")
            Dim weights() As Double = lc.Train(trainData, maxEpochs, seed, 0.0, 0.0)

            Console.WriteLine(vbCrLf & "Best weights found:")
            ShowVector(weights, 3, weights.Length, True)

            Dim trainAccuracy As Double = lc.Accuracy(trainData, weights)
            Console.WriteLine("Prediction accuracy on training data = " & trainAccuracy.ToString("F4"))

            Dim testAccuracy As Double = lc.Accuracy(testData, weights)
            Console.WriteLine("Prediction accuracy on test data = " & testAccuracy.ToString("F4"))

            Console.WriteLine(vbCrLf & "Seeking good L1 weight")
            Dim alpha1 As Double = lc.FindGoodL1Weight(trainData, seed)
            Console.WriteLine("Good L1 weight = " & alpha1.ToString("F3"))

            Console.WriteLine(vbCrLf & "Seeking good L2 weight")
            Dim alpha2 As Double = lc.FindGoodL2Weight(trainData, seed)
            Console.WriteLine("Good L2 weight = " & alpha2.ToString("F3"))

            Console.WriteLine(vbCrLf & "Starting training using L1 regularization, alpha1 = " & alpha1.ToString("F3"))
            weights = lc.Train(trainData, maxEpochs, seed, alpha1, 0.0)

            Console.WriteLine(vbCrLf & "Best weights found:")
            ShowVector(weights, 3, weights.Length, True)

            trainAccuracy = lc.Accuracy(trainData, weights)
            Console.WriteLine("Prediction accuracy on training data = " & trainAccuracy.ToString("F4"))

            testAccuracy = lc.Accuracy(testData, weights)
            Console.WriteLine("Prediction accuracy on test data = " & testAccuracy.ToString("F4"))

            Console.WriteLine(vbCrLf & "Starting training using L2 regularization, alpha2 = " & alpha2.ToString("F3"))
            weights = lc.Train(trainData, maxEpochs, seed, 0.0, alpha2)

            Console.WriteLine(vbCrLf & "Best weights found:")
            ShowVector(weights, 3, weights.Length, True)

            trainAccuracy = lc.Accuracy(trainData, weights)
            Console.WriteLine("Prediction accuracy on training data = " & trainAccuracy.ToString("F4"))

            testAccuracy = lc.Accuracy(testData, weights)
            Console.WriteLine("Prediction accuracy on test data = " & testAccuracy.ToString("F4"))

            Console.WriteLine(vbCrLf & "End Regularization demo" & vbCrLf)
            Console.ReadLine()
        End Sub ' Main

        Private Shared Function MakeAllData(ByVal numFeatures As Integer, ByVal numRows As Integer, ByVal seed As Integer) As Double()()
            Dim rnd As New Random(seed)
            Dim weights(numFeatures) As Double ' inc. b0
            For i  = 0 To weights.Length - 1
                weights(i) = 20.0 * rnd.NextDouble() - 10.0 ' [-10.0 to +10.0]
            Next i

            Dim result(numRows - 1)() As Double ' allocate matrix
            For i = 0 To numRows - 1
                result(i) = New Double(numFeatures) {} ' Y in last column
            Next i

            For i = 0 To numRows - 1 ' for each row
                Dim y As Double = weights(0) ' the b0
                For j = 0 To numFeatures - 1 ' each feature / column except last
                    Dim x As Double = 20.0 * rnd.NextDouble() - 10.0 ' random X in [-10.0, +10.0]
                    result(i)(j) = x ' store x
                    Dim wx As Double = x * weights(j + 1) ' weight * x
                    y += wx ' accumulate to get Y
                    ' now add some noise
                    y += numFeatures * rnd.NextDouble()
                Next j
                If y > numFeatures Then ' because the noise was +, make it harder to be a 1
                    result(i)(numFeatures) = 1.0 ' store y in last column
                Else
                    result(i)(numFeatures) = 0.0
                End If
            Next i
            Console.WriteLine("Data generation weights:")
            ShowVector(weights, 4, 10, True)

            Return result
        End Function ' MakeAllData

        Private Shared Sub MakeTrainTest(ByVal allData()() As Double, ByVal seed As Integer, <System.Runtime.InteropServices.Out()> ByRef trainData()() As Double, <System.Runtime.InteropServices.Out()> ByRef testData()() As Double)
            Dim rnd As New Random(seed)
            Dim totRows As Integer = allData.Length
            Dim numTrainRows As Integer = CInt(Fix(totRows * 0.8)) ' 80% hard-coded
            Dim numTestRows As Integer = totRows - numTrainRows
            trainData = New Double(numTrainRows - 1)() {}
            testData = New Double(numTestRows - 1)() {}

            Dim copy(allData.Length - 1)() As Double ' ref copy of all data
            For i  = 0 To copy.Length - 1
                copy(i) = allData(i)
            Next i

            For i = 0 To copy.Length - 1 ' scramble order
                Dim r As Integer = rnd.Next(i, copy.Length) ' use Fisher-Yates
                Dim tmp() As Double = copy(r)
                copy(r) = copy(i)
                copy(i) = tmp
            Next i
            For i = 0 To numTrainRows - 1
                trainData(i) = copy(i)
            Next i

            For i = 0 To numTestRows - 1
                testData(i) = copy(i + numTrainRows)
            Next i
        End Sub ' MakeTrainTest

        Public Shared Sub ShowData(ByVal data()() As Double, ByVal numRows As Integer, ByVal decimals As Integer, ByVal indices As Boolean)
            Dim len As Integer = data.Length.ToString().Length
            For i = 0 To numRows - 1
                If indices = True Then
                    Console.Write("[" & i.ToString().PadLeft(len) & "]  ")
                End If
                For j = 0 To data(i).Length - 1
                    Dim v As Double = data(i)(j)
                    If v >= 0.0 Then
                        Console.Write(" ") ' '+'
                    End If
                    Console.Write(v.ToString("F" & decimals) & "  ")
                Next j
                Console.WriteLine("")
            Next i
            Console.WriteLine(". . .")
            Dim lastRow As Integer = data.Length - 1
            If indices = True Then
                Console.Write("[" & lastRow.ToString().PadLeft(len) & "]  ")
            End If
            For j = 0 To data(lastRow).Length - 1
                Dim v As Double = data(lastRow)(j)
                If v >= 0.0 Then
                    Console.Write(" ") ' '+'
                End If
                Console.Write(v.ToString("F" & decimals) & "  ")
            Next j
            Console.WriteLine(vbCrLf)
        End Sub

        Public Shared Sub ShowVector(ByVal vector() As Double, ByVal decimals As Integer, ByVal lineLen As Integer, ByVal newLine As Boolean)
            For i = 0 To vector.Length - 1
                If i > 0 AndAlso i Mod lineLen = 0 Then
                    Console.WriteLine("")
                End If
                If vector(i) >= 0 Then
                    Console.Write(" ")
                End If
                Console.Write(vector(i).ToString("F" & decimals) & " ")
            Next i
            If newLine = True Then
                Console.WriteLine("")
            End If
        End Sub

    End Class ' Program

    Public Class LogisticClassifier
        Private numFeatures As Integer ' number of independent variables aka features
        Private weights() As Double ' b0 = constant
        Private rnd As Random

        Public Sub New(ByVal numFeatures As Integer)
            Me.numFeatures = numFeatures ' number features/predictors
            Me.weights = New Double(numFeatures) {} ' [0] = b0 constant
            Me.rnd = New Random(0) ' seed could be a parmeter to ctor
        End Sub

        Public Function FindGoodL1Weight(ByVal trainData()() As Double, ByVal seed As Integer) As Double
            Dim result As Double = 0.0
            Dim bestErr As Double = Double.MaxValue
            Dim currErr As Double = Double.MaxValue
            Dim candidates() As Double = {0.0, 0.001, 0.005, 0.01, 0.02, 0.05, 0.1, 0.15}
            Dim maxEpochs As Integer = 1000

            Dim c As New LogisticClassifier(Me.numFeatures)

            For trial = 0 To candidates.Length - 1
                Dim alpha1 As Double = candidates(trial)
                Dim wts() As Double = c.Train(trainData, maxEpochs, seed, alpha1, 0.0)
                currErr = [Error](trainData, wts, 0.0, 0.0)
                If currErr < bestErr Then
                    bestErr = currErr
                    result = candidates(trial)
                End If
            Next trial
            Return result
        End Function

        Public Function FindGoodL2Weight(ByVal trainData()() As Double, ByVal seed As Integer) As Double
            Dim result As Double = 0.0
            Dim bestErr As Double = Double.MaxValue
            Dim currErr As Double = Double.MaxValue
            Dim candidates() As Double = {0.0, 0.001, 0.005, 0.01, 0.02, 0.05, 0.1, 0.15}
            Dim maxEpochs As Integer = 1000

            Dim c As New LogisticClassifier(Me.numFeatures)

            For trial = 0 To candidates.Length - 1
                Dim alpha2 As Double = candidates(trial)
                Dim wts() As Double = c.Train(trainData, maxEpochs, seed, 0.0, alpha2)
                currErr = [Error](trainData, wts, 0.0, 0.0)
                If currErr < bestErr Then
                    bestErr = currErr
                    result = candidates(trial)
                End If
            Next trial
            Return result
        End Function

        Public Function Train(ByVal trainData()() As Double, ByVal maxEpochs As Integer, ByVal seed As Integer, ByVal alpha1 As Double, ByVal alpha2 As Double) As Double()
            ' use PSO. particle position == LR weights
            Dim numParticles As Integer = 10
            Dim probDeath As Double = 0.005
            Dim [dim] As Integer = Me.numFeatures + 1 ' need one wt for each feature, plus the b0 constant

            Dim rnd As New Random(seed)

            Dim epoch As Integer = 0
            Dim minX As Double = -10.0 ' for each weight. assumes data has been normalized about 0
            Dim maxX As Double = 10.0
            Dim w As Double = 0.729 ' inertia weight
            Dim c1 As Double = 1.49445 ' cognitive/local weight
            Dim c2 As Double = 1.49445 ' social/global weight
            Dim r1, r2 As Double ' cognitive and social randomizations

            Dim swarm(numParticles - 1) As Particle
            ' best solution found by any particle in the swarm. implicit initialization to all 0.0
            Dim bestSwarmPosition([dim] - 1) As Double
            Dim bestSwarmError As Double = Double.MaxValue ' smaller values better

            ' initialize each Particle in the swarm with random positions and velocities
            For i = 0 To swarm.Length - 1
                Dim randomPosition([dim] - 1) As Double
                For j = 0 To randomPosition.Length - 1
                    randomPosition(j) = (maxX - minX) * rnd.NextDouble() + minX
                Next j

                ' randomPosition is a set of weights
                Dim [error] As Double = Me.Error(trainData, randomPosition, alpha1, alpha2)
                Dim randomVelocity([dim] - 1) As Double
                For j = 0 To randomVelocity.Length - 1
                    Dim lo As Double = 0.1 * minX
                    Dim hi As Double = 0.1 * maxX
                    randomVelocity(j) = (hi - lo) * rnd.NextDouble() + lo
                Next j
                swarm(i) = New Particle(randomPosition, [error], randomVelocity, randomPosition, [error]) ' last two are best-position and best-error

                ' does current Particle have global best position/solution?
                If swarm(i).error < bestSwarmError Then
                    bestSwarmError = swarm(i).error
                    swarm(i).position.CopyTo(bestSwarmPosition, 0)
                End If
            Next i
            ' initialization

            ' main PSO algorithm
            Dim sequence(numParticles - 1) As Integer ' process particles in random order
            For i = 0 To sequence.Length - 1
                sequence(i) = i
            Next i

            Do While epoch < maxEpochs
                Dim newVelocity([dim] - 1) As Double ' step 1
                Dim newPosition([dim] - 1) As Double ' step 2
                Dim newError As Double ' step 3

                Shuffle(sequence) ' move particles in random sequence

                For pi = 0 To swarm.Length - 1 ' each Particle (index)
                    Dim i As Integer = sequence(pi)
                    Dim currP As Particle = swarm(i) ' for coding convenience

                    ' 1. compute new velocity
                    For j = 0 To currP.velocity.Length - 1 ' each x value of the velocity
                        r1 = rnd.NextDouble()
                        r2 = rnd.NextDouble()

                        ' velocity depends on old velocity, best position of parrticle, and 
                        ' best position of any particle
                        newVelocity(j) = (w * currP.velocity(j)) + (c1 * r1 * (currP.bestPosition(j) - currP.position(j))) + (c2 * r2 * (bestSwarmPosition(j) - currP.position(j)))
                    Next j

                    newVelocity.CopyTo(currP.velocity, 0)

                    ' 2. use new velocity to compute new position
                    For j = 0 To currP.position.Length - 1
                        newPosition(j) = currP.position(j) + newVelocity(j) ' compute new position
                        If newPosition(j) < minX Then ' keep in range
                            newPosition(j) = minX
                        ElseIf newPosition(j) > maxX Then
                            newPosition(j) = maxX
                        End If
                    Next j

                    newPosition.CopyTo(currP.position, 0)

                    ' 3. use new position to compute new error
                    newError = [Error](trainData, newPosition, alpha1, alpha2)
                    currP.error = newError

                    If newError < currP.bestError Then ' new particle best?
                        newPosition.CopyTo(currP.bestPosition, 0)
                        currP.bestError = newError
                    End If

                    If newError < bestSwarmError Then ' new swarm best?
                        newPosition.CopyTo(bestSwarmPosition, 0)
                        bestSwarmError = newError
                    End If

                    ' 4. optional: does curr particle die?
                    Dim die As Double = rnd.NextDouble()
                    If die < probDeath Then
                        ' new position, leave velocity, update error
                        For j = 0 To currP.position.Length - 1
                            currP.position(j) = (maxX - minX) * rnd.NextDouble() + minX
                        Next j
                        currP.error = [Error](trainData, currP.position, alpha1, alpha2)
                        currP.position.CopyTo(currP.bestPosition, 0)
                        currP.bestError = currP.error

                        If currP.error < bestSwarmError Then ' swarm best by chance?
                            bestSwarmError = currP.error
                            currP.position.CopyTo(bestSwarmPosition, 0)
                        End If
                    End If

                Next pi ' each Particle

                epoch += 1

            Loop ' while

            Dim retResult([dim] - 1) As Double
            Array.Copy(bestSwarmPosition, retResult, retResult.Length)
            Return retResult

        End Function ' Train

        Private Sub Shuffle(ByVal sequence() As Integer)
            For i = 0 To sequence.Length - 1
                Dim r As Integer = rnd.Next(i, sequence.Length)
                Dim tmp As Integer = sequence(r)
                sequence(r) = sequence(i)
                sequence(i) = tmp
            Next i
        End Sub

        Public Function [Error](ByVal trainData()() As Double, ByVal weights() As Double, ByVal alpha1 As Double, ByVal alpha2 As Double) As Double
            ' mean squared error using supplied weights
            ' L1 regularization adds the sum of the absolute values of the weights
            ' L2 regularization adds the sqrt of sum of squared values

            Dim yIndex As Integer = trainData(0).Length - 1 ' y-value (0/1) is last column
            Dim sumSquaredError As Double = 0.0
            For i = 0 To trainData.Length - 1 ' each data
                Dim computed As Double = ComputeOutput(trainData(i), weights)
                Dim desired As Double = trainData(i)(yIndex) ' ex: 0.0 or 1.0
                sumSquaredError += (computed - desired) * (computed - desired)
            Next i

            Dim sumAbsVals As Double = 0.0 ' L1 penalty
            For i = 0 To weights.Length - 1
                sumAbsVals += Math.Abs(weights(i))
            Next i

            Dim sumSquaredVals As Double = 0.0 ' L2 penalty
            For i = 0 To weights.Length - 1
                sumSquaredVals += (weights(i) * weights(i))
            Next i
            'double rootSum = Math.Sqrt(sumSquaredVals);

            Return (sumSquaredError / trainData.Length) + (alpha1 * sumAbsVals) + (alpha2 * sumSquaredVals)
        End Function

        Public Function ComputeOutput(ByVal dataItem() As Double, ByVal weights() As Double) As Double
            Dim z As Double = 0.0

            z += weights(0) ' the b0 constant
            For i = 0 To weights.Length - 2 ' data might include Y
                z += (weights(i + 1) * dataItem(i)) ' skip first weight
            Next i
            Return 1.0 / (1.0 + Math.Exp(-z))
        End Function

        Public Function ComputeDependent(ByVal dataItem() As Double, ByVal weights() As Double) As Integer
            Dim sum As Double = ComputeOutput(dataItem, weights)

            If sum <= 0.5 Then
                Return 0
            Else
                Return 1
            End If
        End Function

        Public Function Accuracy(ByVal trainData()() As Double, ByVal weights() As Double) As Double
            Dim numCorrect As Integer = 0
            Dim numWrong As Integer = 0
            Dim yIndex As Integer = trainData(0).Length - 1
            For i As Integer = 0 To trainData.Length - 1
                Dim computed As Double = ComputeDependent(trainData(i), weights) ' implicit cast
                Dim desired As Double = trainData(i)(yIndex) ' 0.0 or 1.0

                Dim epsilon As Double = 0.0000000001
                If Math.Abs(computed - desired) < epsilon Then
                    numCorrect += 1
                Else
                    numWrong += 1
                End If
            Next i
            Return (numCorrect * 1.0) / (numWrong + numCorrect)
        End Function

        Public Class Particle ' for PSO training
            Public position() As Double ' equivalent to weights
            Public [error] As Double ' measure of fitness
            Public velocity() As Double ' determines new position
            Public bestPosition() As Double ' best found by this Particle
            Public bestError As Double

            Public Sub New(ByVal position() As Double, ByVal [error] As Double, ByVal velocity() As Double, ByVal bestPosition() As Double, ByVal bestError As Double)
                Me.position = New Double(position.Length - 1) {}
                position.CopyTo(Me.position, 0)
                Me.error = [error]
                Me.velocity = New Double(velocity.Length - 1) {}
                velocity.CopyTo(Me.velocity, 0)
                Me.bestPosition = New Double(bestPosition.Length - 1) {}
                bestPosition.CopyTo(Me.bestPosition, 0)
                Me.bestError = bestError
            End Sub
        End Class ' (nested) class Particle

    End Class ' LogisticClassifier

End Namespace ' ns